/*
 *      @(#)JSPositionalSample.java 1.24 99/06/27 17:39:31
 *
 * Copyright (c) 1996-1998 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Sun grants you ("Licensee") a non-exclusive, royalty free, license to use,
 * modify and redistribute this software in source and binary code form,
 * provided that i) this copyright notice and license appear on all copies of
 * the software; and ii) Licensee does not utilize the software in a manner
 * which is disparaging to Sun.
 *
 * This software is provided "AS IS," without a warranty of any kind. ALL
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR
 * NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN AND ITS LICENSORS SHALL NOT BE
 * LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THE SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL SUN OR ITS
 * LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER
 * CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF
 * OR INABILITY TO USE SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control of
 * aircraft, air traffic, aircraft navigation or aircraft communications; or in
 * the design, construction, operation or maintenance of any nuclear
 * facility. Licensee represents and warrants that it will not use or
 * redistribute the Software for such purposes.
 */

/*
 * Java Sound PositionalSample object
 */

package com.sun.j3d.audioengines.javasound;

import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.audioengines.*;

/**
 * The PostionalSample Class defines the data and methods associated with a 
 * PointSound sample played thru the AudioDevice.
 */

class JSPositionalSample extends JSSample
{
    // flag that enables debug prints in calculateDoppler method
    static boolean debugDoppler = false;

    /**
     * For positional and directional sounds, TWO Hae streams or clips
     * are allocated, one each for the left and right channels, played at
     * a different (delayed) time and with a different gain value.
     */
    int     secondIndex = NULL_SAMPLE; 
    /**
     * A third sample for control of reverb of the stream/clip is openned
     * and maintained for all directional/positional sounds.
     * For now, even if no aural attributes (using reverb) are active, 
     * a reverb channel is always started with the other two. A sound could 
     * be started without reverb and then reverb added later, but since there
     * is no way to offset properly into all sounds (considering non-cached
     * and nconsistent rate-changes during playing) this third sound is
     * always allocated and started.
     */
    int     reverbIndex = NULL_SAMPLE; 


    // The transformed position of this sound
    Point3f xformPosition = new Point3f();

    /**
     * Save ear positions transformed into VirtualWorld coords from Head coords
     * These default positions are used when the real values cannot queried
     */
    Point3f  xformCenterEar = new Point3f(0.0f, 0.03f, 0.095f);
    Point3f  xformLeftEar   = new Point3f(-0.09f, -0.03f, 0.095f);
    Point3f  xformRightEar  = new Point3f(0.09f, -0.03f, 0.095f);
    // Z axis in head space - looking into the screen
    Vector3f xformHeadZAxis = new Vector3f(0.0f, 0.0f, -1.0f); // Va

    /**
     * Save vectors from source source position to transformed ear parameters
     */
    Vector3f sourceToCenterEar = new Vector3f();   // Vh
    Vector3f sourceToRightEar = new Vector3f();    // Vf or Vc
    Vector3f sourceToLeftEar = new Vector3f();     // Vf or Vc

    /*
     * Maintain the last locations of sound and head as well as time the
     * sound was last processed
     */
    Point3f  lastXformPosition = new Point3f();
    Point3f  lastXformCenterEar = new Point3f();
    long     lastTimeProcessed = -1;  // denotes not previously processed

    // Size of delta arrays based on this static constant.
    // Value MUST be >= 3 for the Doppler test and correction methods to work.
    static int NUM_DELTAS_KEPT = 5;

    // these delta arrays are circular buffers
    long[]     deltaTime = new long[NUM_DELTAS_KEPT];
    float[]    deltaHeadDistance = new float[NUM_DELTAS_KEPT];
    float[]    deltaSoundDistance = new float[NUM_DELTAS_KEPT];
    float[]    sourceToHead = new float[NUM_DELTAS_KEPT];
    int        curIndex = 0;
    boolean    deltaArraysFilled = false;

    /*
     * Allow changes in Doppler rate only small incremental values otherwise
     * you hear skips in the pitch of a sound during playback.
     * When playback is faster, allow delta changes:
     *   (diff in Factor for octave (1.0))/(12 1/2-steps))*(1/4) of half-step 
     * When playback is slower, allow delta changes:
     *   (diff in Factor for octave (0.5))/(12 1/2-steps))*(1/4) of half-step 
     */
    double   lastRequestedDopplerRateRatio = -1.0f;
    double   lastActualDopplerRateRatio    = -1.0f;
    /*
     * KLUDGE - limit the first calculated ratio (before user-defined
     *          frequency or roll off parameters factored in.
     *          Force to a minor 3rd (1/4 of an octive) which shifted up
     *          is the ratio 0.25, and down is 0.125.
     */
    static   double maxRatio                 = 1.25f;   // 1/4 an octive higher
    static   double minRatio                 = 0.875;   // 1/4 an octive lower
    /*
     * denotes movement of sound away or towards listener
     */
    static   int    TOWARDS   =  1;
    static   int    NO_CHANGE =  0;
    static   int    AWAY      = -1;

    /*
     * Process request for Filtering fields
     */  
    boolean   filterFlag = false;
    float     filterFreq = -1.0f;


    /*
     * Construct a new audio device Sample object
     */  
    public JSPositionalSample() {
        super();
        if (debugFlag)
            debugPrint("JSPositionalSample constructor");
    }

    public void  setSecondIndex(int index) {
        secondIndex = index;
    }

    public int   getSecondIndex() {
        return secondIndex;
    }

    public void  setReverbIndex(int index) {
        reverbIndex = index;
    }

    public int   getReverbIndex() {
        return reverbIndex;
    }

    /*
     * Process request for Filtering fields
     */  
    public boolean  getFilterFlag() {
        return filterFlag;
    }
    public float  getFilterFreq() {
        return filterFreq;
    }

    
    /**
     * Clears the fields associated with sample data for this sound, and 
     * frees any device specific data associated with this sample.
     */  
    public void clear() {
        if (debugFlag)
            debugPrint("JSPositionalSample.clear() enter");
        super.clear();
        xformPosition.set(0.0f, 0.03f, 0.095f);
        xformCenterEar.set(0.0f, 0.03f, 0.095f);
        xformLeftEar.set(-0.09f, -0.03f, 0.095f);
        xformRightEar.set(0.09f, -0.03f, 0.095f);
        // Z axis in head space - looking into the screen
        xformHeadZAxis.set(0.0f, 0.0f, -1.0f); // Va
        sourceToCenterEar.set(0.0f, 0.0f, 0.0f);   // Vh
        sourceToRightEar.set(0.0f, 0.0f, 0.0f);    // Vf or Vc
        sourceToLeftEar.set(0.0f, 0.0f, 0.0f);     // Vf or Vc
        lastXformPosition.set(0.0f, 0.0f, 0.0f);
        lastXformCenterEar.set(0.0f, 0.0f, 0.0f);
        lastTimeProcessed = -1;  // denotes not previously processed
        for (int i=0; i<NUM_DELTAS_KEPT; i++) {
            deltaTime[i] = 0L;
            deltaHeadDistance[i] = 0.0f; 
            deltaSoundDistance[i] = 0.0f; 
        }
        curIndex = 0;
        deltaArraysFilled = false;
        rateRatio = 1.0f;
        if (debugFlag)
            debugPrint("JSPositionalSample.clear() exit");
    }

    void setXformedPosition() {
        if (debugFlag)
            debugPrint("*** setXformedPosition");
        if (!getVwXfrmFlag()) {
            if (debugFlag)
                debugPrint("    Transform NOT set yet, so position NOT set");
            return;  // can't set transformed position until transform gotten
        }
        /**
         * Update the last* fields first...
         */ 
        long currentTime = System.currentTimeMillis(); 
        if (lastTimeProcessed > 0) {
            deltaTime[curIndex] = currentTime - lastTimeProcessed; 
            lastXformPosition.x = xformPosition.x;
            lastXformPosition.y = xformPosition.y;
            lastXformPosition.z = xformPosition.z;
            if (debugFlag) {
                debugPrint("           deltaTime set to " + deltaTime[curIndex]);
                debugPrint("           lastXform(sound)Position = " + 
                       lastXformPosition.x + ", " + lastXformPosition.y + 
                       ", " + lastXformPosition.z);
            }
        }
        else {
            deltaTime[curIndex] = 0;
            if (debugFlag)
                debugPrint("           No LAST values yet!");
        }
	vworldXfrm.transform(position, xformPosition);
        if (debugFlag)
            debugPrint("           xform(sound)Position = " + xformPosition.x +
                       ", " + xformPosition.y + ", " + xformPosition.z);
        lastTimeProcessed = currentTime;
    } 

    /**
     * Set Doppler effect Rate 
     *
     * Calculate the rate of change in for the head and sound
     * between the two time stamps (last two times position or
     * VirtualWorld transform was updated).
     * First determine if the head and sound source are moving 
     * towards each other (distance between them is decreasing), 
     * moving away from each other (distance between them is
     * increasing), or no change (distance is the same, not moving
     * or moving the same speed/direction).
     * The following equation is used for determining the change in frequency -
     *     If there has been a change in the distance between the head and sound:
     *
     *         f' =  f * frequencyScaleFactor * velocityRatio
     *
     *     For no change in the distance bewteen head and sound, velocityRatio is:
     *
     *         f' =  f
     *
     *     For head and sound moving towards each other, velocityRatio (> 1.0) is:
     *
     *         |  speedOfSound*rollOff  +   velocityOfHead*velocityScaleFactor  |
     *         |  ------------------------------------------------------------- |
     *         |  speedOfSound*rollOff  -  velocityOfSource*velocityScaleFactor |
     * 
     *     For head and sound moving away from each other, velocityRatio (< 1.0) is:
     * 
     *         |  speedOfSound*rollOff  -   velocityOfHead*velocityScaleFactor  |
     *         |  ------------------------------------------------------------- |
     *         |  speedOfSound*rollOff  +  velocityOfSource*velocityScaleFactor |
     *
     * where frequencyScaleFactor, rollOff, velocityScaleFactor all come from
     * the active AuralAttributes parameters.
     * The following special cases must be test for AuralAttribute parameters:
     *   rolloff  
     *       Value MUST be > zero for any sound to be heard!
     *       If value is zero, all sounds affected by AuralAttribute region are silent.
     *   velocityScaleFactor 
     *      Value MUST be > zero for any sound to be heard!
     *      If value is zero, all sounds affected by AuralAttribute region are paused.
     *   frequencyScaleFactor 
     *      Value of zero disables Doppler calculations:
     *         Sfreq' = Sfreq * frequencyScaleFactor
     *
     * The 'velocityOfHead' is actually velocity of head RELATIVE to the 
     * sound source.  This is calculated as velocity of head (over
     * delta time) times cos(h) where h is the angle between the
     * vector between the last to current head positions and the vector 
     * from the last sound source to last head positions.
     * The 'velocityOfSound' is actually velocity of sound RELATIVE to the 
     * listener's position.  This is calculated as velocity of source (over
     * delta time) times cos(s) where s is the angle between the vector
     * from last to current sound source positions and vector from last sound 
     * source to last head positions.
     *
     * This rate is passed to device drive as a change to playback sample
     * rate, in this case the frequency need not be known.
     *
     * Return value of zero denotes no change
     * Return value of -1 denotes ERROR

@@@@@@
NOTES
@@@@@@
%%% TODO: need to consider the following:
    Throw out deltaTime values
    Limit upward shift by 3rd?
    Limit downward shifts by something much smaller?
    If direction shift, ignor slow incremental shift trick!
    If movement STOPs - how would I know? - it would be nice to stop
      gradual shift in shifting.
    If change from towards to away, make shift occur faster!!!

     */
    float calculateDoppler(AuralParameters attribs) {
        double sampleRateRatio = 1.0;
        double headVelocity = 0.0;              // in milliseconds
        double soundVelocity = 0.0;             // in milliseconds
        double distanceSourceToHead = 0.0;      // in meters
        double lastDistanceSourceToHead = 0.0;  // in meters
        float  speedOfSound = attribs.SPEED_OF_SOUND;
        double numerator = 1.0;
        double denominator = 1.0;
        int    direction   = 0; // denotes movement away or towards listener
        int    lastIndex = curIndex-1;
        int calcIndex = (curIndex-(NUM_DELTAS_KEPT/2)) % NUM_DELTAS_KEPT;

        distanceSourceToHead = xformPosition.distance(xformCenterEar);

        if (lastIndex < 0)
            lastIndex = NUM_DELTAS_KEPT - 1;  // reset to top of circular buffer
        lastDistanceSourceToHead = sourceToHead[lastIndex];

        calcIndex = (curIndex-(NUM_DELTAS_KEPT/2)) % NUM_DELTAS_KEPT;
        if (calcIndex < 0)
            calcIndex += NUM_DELTAS_KEPT;

        if (debugDoppler) {
            debugPrint(true, "JSPositionalSample.calculateDoppler - distances: " +
                           "current,last = " + distanceSourceToHead + ", " +
                           lastDistanceSourceToHead );
            debugPrint(true, "                                      " +
                "current position = " +
                xformPosition.x + ", " + xformPosition.y +
                ", " + xformPosition.z);
            debugPrint(true, "                                      " +
                "current ear = " +
                xformCenterEar.x + ", " + xformCenterEar.y +
                ", " + xformCenterEar.z);
            debugPrint(true, "                                      " +
                "last position = " +
                lastXformPosition.x + ", " + lastXformPosition.y +
                ", " + lastXformPosition.z);
            debugPrint(true, "                                      " +
                "last ear = " +
                lastXformCenterEar.x + ", " + lastXformCenterEar.y +
                ", " + lastXformCenterEar.z);
        }
        if ( (distanceSourceToHead == lastDistanceSourceToHead)  ||
             (deltaTime[curIndex] <= 0L) ) {
            // %%% TODO Use some EPSilion to do 'equals' test against
            // Deltas and distances are not kept for this pass
            if (debugDoppler)
                debugPrint(true, "JSPositionalSample.calculateDoppler - " +
                           "distance diff = 0, dopplerRatio set to zero");
            direction = NO_CHANGE;
            sampleRateRatio = 0.0;
        }
        else {
            deltaSoundDistance[curIndex] = 
                    xformPosition.distance(lastXformPosition);
            deltaHeadDistance[curIndex] = 
                    xformCenterEar.distance(lastXformCenterEar);
            sourceToHead[curIndex] = (float)distanceSourceToHead;
            boolean goodIndex = true;  
            if (!deltaArraysFilled) {
                if (curIndex < (NUM_DELTAS_KEPT-1)) {
                   direction = NO_CHANGE;
                   sampleRateRatio = 0.0;
                   curIndex ++;
                }
                else {
                   deltaArraysFilled = true;
                   curIndex = 0;
                }
            }
            // the below "if" test is needed rather than just an "else"
            // since deltaArraysFilled could have changed in above if-block
            if (deltaArraysFilled) { // at least NUM_DELTAS_KEPT elements have been stored
                // at least NUM_DELTAS_KEPT elements have been stored
                // %%% For now test only time elements
                /*** EARLY ATTEMPT 
                int badIndices = 0;  // bit mask denoting which elements in delta array
                badIndices = compareDelta(deltaTime);
                if (badIndices > 0)  {
                    // used shuffleElements(badIndices, *) for each delta array
                }
                ***/
                goodIndex = compareDelta(calcIndex, deltaTime);
                if (!goodIndex) {
                    // thru out calcIndex and shift remaining elements 
                    removeElement(calcIndex, deltaTime);
// %%% TODO if I don't need to check the delta values of distances I don't need to
//          keep distance arrays
                    removeElement(calcIndex, deltaHeadDistance);
                    removeElement(calcIndex, deltaSoundDistance);
                    // don't increment curIndex
                }
                else {
                    // all the numbers in delta arrays are keepers
                    soundVelocity = 
                        deltaSoundDistance[calcIndex]/deltaTime[calcIndex];
                    headVelocity = 
                        deltaHeadDistance[calcIndex]/deltaTime[calcIndex];
                    // Now apply the cos(s) and cos(h), described in the
                    // method's header above, using dot product equations.
                    // %%% QUESTION is using Last distance from Source to Head 
                    //     correct, or should Current distance be used?
                    if (soundVelocity != 0.0f &&
                          deltaSoundDistance[curIndex] > 0.0f &&
                          sourceToHead[lastIndex] > 0.0f) {
                        if (debugDoppler) debugPrint(true,
                          "                                      " + 
                          "soundVelocity before *cos(s) = " + soundVelocity);
                        soundVelocity *= 
                            ( (xformPosition.x * lastXformPosition.x) +
                              (xformPosition.y * lastXformPosition.y) +
                              (xformPosition.z * lastXformPosition.z) ) /
                            ( deltaSoundDistance[curIndex] *
                              sourceToHead[lastIndex] ) ;
                        if (debugDoppler) debugPrint(true,
                          "                                      " + 
                          "soundVelocity after *cos(s) = " + soundVelocity);
                    }
                    if (headVelocity != 0.0f && 
                          deltaHeadDistance[calcIndex] > 0.0f &&
                          sourceToHead[lastIndex] > 0.0f ) {
                        if (debugDoppler) debugPrint(true,
                          "                                      " + 
                          "headVelocity before *cos(h) = " + headVelocity);
                        headVelocity *=  deltaSoundDistance[curIndex] * 
                            ( (xformCenterEar.x * lastXformCenterEar.x ) +
                              (xformCenterEar.y * lastXformCenterEar.y ) +
                              (xformCenterEar.z * lastXformCenterEar.z ) ) /
                            ( deltaHeadDistance[calcIndex] * 
                              sourceToHead[lastIndex] ) ;
                        if (debugDoppler) debugPrint(true,
                          "                                      " + 
                          "headVelocity after *cos(h) = " + headVelocity);
                    }
                    if (debugDoppler) {
                        debugPrint(true, "                                      " +
                            "delta time = " + deltaTime[calcIndex] );
                        debugPrint(true, "                                      " +
                            "soundPosition delta = " + 
                            xformPosition.distance(lastXformPosition));
                        debugPrint(true, "                                      " +
                            "soundVelocity = " + soundVelocity);
                        debugPrint(true, "                                      " +
                            "headPosition delta = " + 
                            xformCenterEar.distance(lastXformCenterEar));
                        debugPrint(true, "                                      " +
                            "headVelocity = " + headVelocity);
                    }
                    if (distanceSourceToHead < lastDistanceSourceToHead) {
                        // sound and head moving towards each other
                        if (debugDoppler)
                            debugPrint(true, "                                      " +
                             "moving TOWARDS...");
                        direction = TOWARDS;
                        numerator = speedOfSound + headVelocity;
                        denominator = speedOfSound - soundVelocity;
                    }
                    else { 
                        // sound and head moving away from each other
                        //    note: no change in distance case covered above
                        if (debugDoppler)
                            debugPrint(true, "                                      " +
                             "moving AWAY...");
                        direction = AWAY;
                        numerator = speedOfSound - headVelocity;
                        denominator = speedOfSound + soundVelocity;
                    }

                    double unscaledRatio = numerator / denominator;
                    double limitingFactor = 1.0f;  // unchanged by default
                    if (unscaledRatio > maxRatio || unscaledRatio < minRatio) {
                        limitingFactor = maxRatio / unscaledRatio;
                        if (debugDoppler)
                            debugPrintln(true,
                              "                                      " +
                              "limiting factor is = " + limitingFactor);
                    }

                    if (attribs != null) {
        
                        float rolloff = attribs.getRolloff();
                        float velocityScaleFactor = attribs.getVelocityScaleFactor();
                        if (rolloff != 1.0f) {
                            speedOfSound *=  rolloff;
                            if (debugDoppler)
                                debugPrint(true, "                                      " +
                                    "attrib rollof = " + rolloff);
                        }
                        if (debugDoppler)
                            debugPrint(true, "                                      " +
                                    "speed of sound = " + speedOfSound);
                        if (velocityScaleFactor != 1.0f) {
                            soundVelocity *= velocityScaleFactor;
                            headVelocity *= velocityScaleFactor;
                            if (debugDoppler) {
                                debugPrint(true, "                                      " +
                            "attrib velocity scale factor = " + 
                                     velocityScaleFactor );
                                debugPrint(true, "                                      " +
                                    "new soundVelocity = " + soundVelocity);
                                debugPrint(true, "                                      " +
                                    "new headVelocity = " + headVelocity);
                            }
                        }
                        /*
                         * Now that speed of sound & velocities have been scaled
                         * by their relative factors from the soundscape attri-
                         * butes * recalculate numerator and denominator
                         */
                        if (direction == TOWARDS) {
                            // sound and head moving towards each other
                            numerator = speedOfSound + headVelocity;
                            denominator = speedOfSound - soundVelocity;
                        }
                        else if (direction == AWAY) {
                            // sound and head moving away from each other
                            numerator = speedOfSound - headVelocity;
                            denominator = speedOfSound + soundVelocity;
                        }
                    }
                    if (numerator <= 0.0) {
                        if (debugDoppler)
                            debugPrint(true, "JSPositionalSample.calculateDoppler: " +
                                       "BOOM!! - velocity of head > speed of sound");
                        sampleRateRatio = -1.0;
                    }
                    else if (denominator <= 0.0) {
                        if (debugDoppler)
                            debugPrint(true, "JSPositionalSample.calculateDoppler: " + 
                                       "BOOM!! - velocity of sound source negative");
                        sampleRateRatio = -1.0;
                    }
                    else {
                        if (debugDoppler)
                            debugPrint(true, "JSPositionalSample.calculateDoppler: " +
                                       "numerator = " + numerator +
                                       ", denominator = " + denominator );
                        sampleRateRatio = numerator / denominator;
                        if (limitingFactor != 1.0f)
                            sampleRateRatio *= limitingFactor;
                    }
                } // end of block when there is a change in distance to sound
                curIndex++;
                if (curIndex == NUM_DELTAS_KEPT) 
                      curIndex = 0;
            }   
        }
        if (debugDoppler)
            debugPrintln(true, "JSPositionalSample.calculateDoppler: " +
                    "sampleRateRatio before adjustment = " + sampleRateRatio);

        // 
        // 
        // Gradually change rate scale factor
        // 
        // If the rate change is too great suddenly, it sounds like a
        // jump, so we need to change gradually over time.
        // Since an octive delta change up is 2.0 but down is 0.5, forced
        // "max" rate of change is different for both.
        // 
        // %%% QUESTION what should max delta changes allowed for rate changes be
        //     Tried  0.0104, 0.0208 but these values were too large.
        //     Even 0.0052 caused noticible shifts in some tests
        //        thought this was 1/32 of 1/2 step

// %%% PROBLEM-BUG
// Wildly "garbled" sounds apprear to be caused by unequal delta changes in
// time verses distance changes.
// For examples from one run of the test EnableJSSilent we got delta times
//     that varied wildly from delta xformed position (distance)
//         delta time = 45, soundPosition delta = 0.0735179
//         delta time = 13, soundPosition delta = 0.243152
//         delta time = 13, soundPosition delta = 0.56545514
//         delta time = 27, soundPosition delta = 0.4523887
//         delta time = 16, soundPosition delta = 0.38452744
//         delta time = 44, soundPosition delta = 0.3505776
//         delta time = 29, soundPosition delta = 0.44108152
//         delta time = 13, soundPosition delta = 0.66721845
//         delta time = 17, soundPosition delta = 0.3788757
// %%% IDEA What if I make change down (as sound moves away) change FASTER
//     since is now moving away from listener.  Try 0.00104 rather than 0.00026
        double maxDopplerRateChangeDown = 0.00104f;  // TRY larger than UpRate
        double maxDopplerRateChangeUp   = 0.00052f;
        double diff;
        if (sampleRateRatio < 0.0) {
            // error 
            if ( lastRequestedDopplerRateRatio < 0.0 || 
                 lastActualDopplerRateRatio < 0.0 ) {
                lastRequestedDopplerRateRatio = -1.0;
                lastActualDopplerRateRatio = -1.0;
                // leave sampleRateRatio -1
            }
            else if ( lastActualDopplerRateRatio >= maxRatio ) {
                // leave sampleRateRatio -1 so sound can be made silent
            }
            else {
                diff = lastRequestedDopplerRateRatio - 
                             lastActualDopplerRateRatio;
                if (diff > 0.0) {
                    // inch up towards lastRequestedDopplerRateRatio
                    if (diff >= maxDopplerRateChangeUp)
                         sampleRateRatio = lastActualDopplerRateRatio +
                                maxDopplerRateChangeUp;
                    else
                         sampleRateRatio = lastActualDopplerRateRatio + diff;
                }
                else if (diff < 0.0) {
                    // inch down towards lastRequestedDopplerRateRatio
                    if ((-diff) >= maxDopplerRateChangeDown)
                         sampleRateRatio = lastActualDopplerRateRatio -
                                maxDopplerRateChangeDown;
                    else
                         sampleRateRatio = lastActualDopplerRateRatio + diff;
                }
                // otherwise no difference between requested and actual ratios
                //     so do nothing and leave sampleRateRatio -1

                lastActualDopplerRateRatio = (double)sampleRateRatio;
            }
            // Don't change the lastRequestedDopplerRateRatio
        } // sampleRateRatio was -1

        else if (sampleRateRatio == 0.0) {   // direction = NO_CHANGE
            // no change between sound and head this pass
            if ( lastRequestedDopplerRateRatio < 0.0 || 
                 lastActualDopplerRateRatio < 0.0 ) {
                lastRequestedDopplerRateRatio = 1.0;
                lastActualDopplerRateRatio = 1.0;
                // leave sampleRateRatio zero
            }
            else if ( lastActualDopplerRateRatio >= maxRatio ) {
                // leave sampleRateRatio 0 so no change is applied
            }
            else { 
                // Caused very wobbly playback
                // maxDopplerRateChangeDown *= lastActualDopplerRateRatio;
                // maxDopplerRateChangeUp   *= lastActualDopplerRateRatio;
                diff = lastRequestedDopplerRateRatio - 
                             lastActualDopplerRateRatio;
                if (diff > 0.0) {
                    // inch up towards lastRequestedDopplerRateRatio
                    if (diff >= maxDopplerRateChangeUp)
                         sampleRateRatio = lastActualDopplerRateRatio +
                                maxDopplerRateChangeUp;
                    else
                         sampleRateRatio = lastActualDopplerRateRatio + diff;
                    lastActualDopplerRateRatio = sampleRateRatio;
                }
                else if (diff < 0.0) {
                    // inch down towards lastRequestedDopplerRateRatio
                    if ((-diff) >= maxDopplerRateChangeDown)
                         sampleRateRatio = lastActualDopplerRateRatio -
                                maxDopplerRateChangeDown;
                    else
                         sampleRateRatio = lastActualDopplerRateRatio + diff;
                    lastActualDopplerRateRatio = sampleRateRatio;
                }
                // otherwise no difference between requested and actual ratios
                //     so do nothing and leave sampleRateRatio zero
            }
            // Don't change the lastRequestedDopplerRateRatio for any case
        } // sampleRateRatio is zero

        else if (sampleRateRatio > 0.0) {
            if ( lastRequestedDopplerRateRatio < 0 || 
                 lastActualDopplerRateRatio < 0 ) {
                if (debugDoppler) {
                    debugPrint(true, "                   calculateDoppler: " +
                        "lastRequested or ActualDopplerRateRatio < 0");
                    debugPrint(true, "                   calculateDoppler: " +
                        "    just set lastRequested & Actual to sampleRateRatio");
                    debugPrint(true, "                   calculateDoppler: " +
                        "    sampleRateRatio left unchanged");
                }
                lastRequestedDopplerRateRatio = sampleRateRatio;
                lastActualDopplerRateRatio = sampleRateRatio;
                // leave sampleRateRatio unchanged
            }
            else {  // we have valid last Doppler rate values
                // Got very wobbly playback
                // maxDopplerRateChangeDown *= lastActualDopplerRateRatio;
                // maxDopplerRateChangeUp   *= lastActualDopplerRateRatio;
                diff = sampleRateRatio - lastActualDopplerRateRatio;
                if (debugDoppler) {
                    debugPrint(true, "                   calculateDoppler: " +
                        "diff between sampleRatio & lastActualDopplerRateRatio" +
                        "("+lastActualDopplerRateRatio+") = " + diff); 
                }
                // lastActualDopplerRateRatio should not be > maxRatio
                if (diff > 0.0) {
                    // go up towards the current sampleRateRatio
                    lastRequestedDopplerRateRatio = sampleRateRatio;
                    if (debugDoppler) {
                        debugPrint(true, "                   calculateDoppler: " +
                            "lastRequestedDopplerRateRatio = sampleRateRatio"); 
                    }
                    if (diff >= maxDopplerRateChangeUp) {
                        // inch up towards sampleRateRatio
                        sampleRateRatio = lastActualDopplerRateRatio +
                                    maxDopplerRateChangeUp;
                        if (debugDoppler) {
                            debugPrint("                   calculateDoppler: " +
                                "diff >= maxDopplerRateChangeUp so ");
                            debugPrint(true, "                   calculateDoppler: " +
                                "    sampleRateRatio incremented up by max");
                        }
                    }
                    /* 
                     * otherwise delta change is within tolerance
                     *    so use sampleRateRatio as calculated w/out change
                     */
                    else {
                        if (debugDoppler) {
                            debugPrint(true, "                   calculateDoppler: " +
                                "diff < maxDopplerRateChangeUp so ");
                            debugPrint(true, "                   calculateDoppler: " +
                                "    sampleRateRatio used without change");
                        }
                    }
                    lastActualDopplerRateRatio = sampleRateRatio;
                    if (debugDoppler) {
                        debugPrint(true, "                   calculateDoppler: " +
                            "lastActualDopplerRateRatio = sampleRateRatio"); 
                    }
                }
                else if (diff < 0.0) {
                    // go down towards current sampleRateRatio
                    lastRequestedDopplerRateRatio = sampleRateRatio;
                    if (debugDoppler) {
                        debugPrint(true, "                   calculateDoppler: " +
                            "lastRequestedDopplerRateRatio = sampleRateRatio"); 
                    }
                    if ((-diff) >= maxDopplerRateChangeDown) {
                        // inch down towards sampleRateRatio
                        sampleRateRatio = lastActualDopplerRateRatio -
                                    maxDopplerRateChangeDown;
                        if (debugDoppler) {
                            debugPrint(true, "                   calculateDoppler: " +
                                "-(diff) >= maxDopplerRateChangeUp so ");
                            debugPrint(true, "                   calculateDoppler: " +
                                "    sampleRateRatio incremented down by max ");
                        }
                    }
                    /*
                     * otherwise negitive delta change is within tolerance so
                     * use sampleRateRatio as calculated w/out change
                     */
                    else {
                        if (debugDoppler) {
                            debugPrint(true, "                   calculateDoppler: " +
                                "-(diff) < maxDopplerRateChangeUp so ");
                            debugPrint(true, "                   calculateDoppler: " +
                                "    sampleRateRatio used without change");
                        }
                    }
                    lastActualDopplerRateRatio = sampleRateRatio;
                    if (debugDoppler) {
                        debugPrint(true, "                   calculateDoppler: " +
                            "lastActualDopplerRateRatio = sampleRateRatio"); 
                    }
                }
                else {
                    // no difference between lastActualDopplerRateRatio and 
                    // sampleRateRatio.
                    if (debugDoppler)
                        debugPrint(true, "                   calculateDoppler: " +
                            "no diff between lastActualDopplerRateRatio "+
                            " & sampleRateRatio");
                    // Inch up towards last Requested Rate
                    // Don't change lastRequestedDopplerRateRatio
                    diff = sampleRateRatio - lastRequestedDopplerRateRatio;
                    if (diff > 0.0) {
                        // inch up towards lastActualDopplerRateRatio
                        if (diff >= maxDopplerRateChangeUp) {
                            sampleRateRatio = lastActualDopplerRateRatio +
                                    maxDopplerRateChangeUp;
                            if (debugDoppler) {
                                debugPrint(true, "                   calculateDoppler:"+
                                    "diff >= maxDopplerRateChangeUp so ");
                                debugPrint(true, "                   calculateDoppler:"+
                                    "    sampleRateRatio incremented up by max");
                            }
                        }
                        else {
                            sampleRateRatio = lastActualDopplerRateRatio + diff;
                            if (debugDoppler) {
                                debugPrint(true, "                   calculateDoppler:"+
                                    " diff2 < maxDopplerRateChangeUp so ");
                                debugPrint(true, "                                    "+
                                    "      sampleRateRatio inched up by " + diff);
                            }
                        }
                        lastActualDopplerRateRatio = sampleRateRatio;
                        if (debugDoppler) {
                            debugPrint(true, "                   calculateDoppler: " +
                                "lastActualDopplerRateRatio = sampleRateRatio"); 
                        }
                    }
                    else if (diff < 0.0) {
                        // inch down towards lastActualDopplerRateRatio
                        if ((-diff) >= maxDopplerRateChangeDown) {
                            sampleRateRatio = lastActualDopplerRateRatio -
                                    maxDopplerRateChangeDown;
                            if (debugDoppler) {
                                debugPrint(true, "                   calculateDoppler:"+
                                    "-(diff) >= maxDopplerRateChangeUp so ");
                                debugPrint(true, "                   calculateDoppler:"+
                                    "    sampleRateRatio incremented down by max");
                            }
                        }
                        else {
                            sampleRateRatio = lastActualDopplerRateRatio + diff;
                            if (debugDoppler) {
                                debugPrint(true, "                   calculateDoppler:"+
                                    " -(diff2) < maxDopplerRateChangeUp so ");
                                debugPrint(true, "                                    "+
                                    "      sampleRateRatio inched down by " + diff);
                            }
                        }
                        lastActualDopplerRateRatio = sampleRateRatio;
                        if (debugDoppler) {
                            debugPrint(true, "                   calculateDoppler: " +
                                "lastActualDopplerRateRatio = sampleRateRatio"); 
                        }

                    }
                    else {
                        // no difference between current sample rate and last
                        //     requested ratio so set sampleRateRatio to zero
                        sampleRateRatio = 0.0;
                        if (debugDoppler) {
                            debugPrint(true, "                   calculateDoppler: " +
                                "no diff(2) between sampleRateRatio & " +
                                "lastRequestedDopplerRatio so ");
                            debugPrint(true, "                                     " +
                                "sampleRateRatio set to zero");
                            debugPrint(true, "                   calculateDoppler: " +
                                "neither lastActual or RequestDopplerRateRatio " +
                                "changed");
                        }
                    }
                }
            }
        } 

        else {
            // This shouldn't be executed
            if (debugDoppler)
                debugPrint(true, "JSPositionalSample.calculateDoppler: " +
                    "invalid sampleRateRatio");
        }

        if (sampleRateRatio > maxRatio)
            sampleRateRatio = maxRatio;
        if (debugDoppler) {
            debugPrintln(true, "JSPositionalSample.calculateDoppler: " +
                    "sampleRateRatio after adjustment = " + sampleRateRatio);
        }
        return (float)sampleRateRatio;
    }
  
    // Assumes NUM_DELTAS_KEPT and thus length of array is >= 3
    int  compareDeltas(long[] array) {
        int bitMask = 0;
        float average = 0.0f;
        int   arrayLen = array.length;
        float[] diff = new float[arrayLen];
        for (int i=0; i<arrayLen; i++) {
            average += (float)array[i];
            if (i < (arrayLen-1))
                diff[i] =  (float)(Math.abs(array[i] - array[i+1]));
            else
                diff[arrayLen-1] =  (float)(Math.abs(array[i] - array[0]));
        }
        average /= arrayLen;
        float epsilon = average * 2.0f;  // %%% KLUDGE first pass at correct value
        for (int i=1; i<(arrayLen-1); i++) {
            if (diff[i-1] > epsilon) {
                if (diff[i] > epsilon) {
                   // element i is a spike that should be thrown out
                   bitMask += 1<<i;
                   if (debugDoppler) 
                        debugPrintln(true,"element " + i + " is bad: " + array[i-1] + 
                                     ", " + array[i]  + ", " + array[i+1] );
                }
                // otherwise keep it
            }
        }
        return bitMask;
    }
    int  compareDeltas(float[] array) {
        int bitMask = 0;
// %%% TODO put contents of other compareDeltas() into this method
        return bitMask;
    }

    // Returns false if deltas between values around value[index] is 'bad'
    boolean  compareDelta(int index, long[] array) {
        int bitMask = 0;
        float average = 0.0f;
        int   arrayLen = array.length;
        float[] diff = new float[2];
        for (int i=0; i<arrayLen; i++) {
            average += (float)array[i];
        }
        average /= arrayLen;
        if (index == 0) {
            diff[0] = (float)(Math.abs(array[0] - array[arrayLen-1]));
            diff[1] = (float)(Math.abs(array[1] - array[0]));
        }
        else if (index == (arrayLen-1)) {
            diff[0] = (float)(Math.abs(array[index] - array[index - 1]));
            diff[1] = (float)(Math.abs(array[0] - array[index]));
        }
        else { // index is between 0 and (arrayLen-1)
            diff[0] = (float)(Math.abs(array[index] - array[index - 1]));
            diff[1] = (float)(Math.abs(array[index+1] - array[index]));
        }
        float epsilon = average * 2.0f;  // %%% KLUDGE first pass at correct value
        if (diff[0] > epsilon) {
            if (diff[1] > epsilon) {
                // element i is a spike that should be thrown out
                if (debugDoppler) {
                      debugPrint(true,"calculateDoppler: Spike in elements around " + index + " found -");
                      debugPrint(true,"     diffs are: " + diff[0] + ", " + diff[1]);
                      debugPrint(true,"     while epsilon is " + epsilon);
                }
                return false;
            }
        }
        return true;
    }

    boolean  compareDelta(int index, float[] array) {
// %%% TODO put contents of other compareDelta() into this method
//          no casting to float necessary
        return true;
    }

    void  removeElement(int index, long[] array) {
        int arrayLen = array.length;
        int numGoodElements = arrayLen;
        for (int i=0, j=index, k=index+1; i<array.length; i++, k++, j++) {
            if (j==array.length)
                j = 0;
            if (k==array.length)
                k = 0;
            array[j] = array[k];
        } 
    }
    void  removeElement(int index, float[] array) {
        int arrayLen = array.length;
        int numGoodElements = arrayLen;
        for (int i=0, j=index, k=index+1; i<array.length; i++, k++, j++) {
            if (j==array.length)
                j = 0;
            if (k==array.length)
                k = 0;
            array[j] = array[k];
        } 
    }

    int  shuffleElements(int bitMask, long[] array) {
        int arrayLen = array.length;
        int numGoodElements = arrayLen;
        if (bitMask > 0)
        for (int i=0, j=0; i<array.length; i++, j++) {
            if ((bitMask&(i<<i)) > 0) {
                array[j] = array[i+1];
                numGoodElements--;
                j--;
            }
        } 
        return numGoodElements;
    }
    int shuffleElements(int bitMask, float[] array) {
        // code for this method IDENTICAL to long version of same method above
        int arrayLen = array.length;
        int numGoodElements = arrayLen;
        if (bitMask > 0)
        for (int i=0, j=0; i<array.length; i++, j++) {
            if ((bitMask&(i<<i)) > 0) {
                array[j] = array[i+1];
                numGoodElements--;
                j--;
            }
        } 
        return numGoodElements;
    }

    void updateEar(int dirtyFlags, View view) {
        if (debugFlag)
            debugPrint("*** updateEar fields");
        /*
         * Transform ear position (from Head) into Virtual World Coord space
         */
        // %%% TODO check dirty flags coming in
        //     For now just recalculate ear positions by forcing earsXformed false
        boolean  earsXformed = false;  
        Point3d  earPosition = new Point3d();   // temporary double Point

        if (!earsXformed) { 
              if (view != null) {
                PhysicalBody body = view.getPhysicalBody();
                if (body != null) {
                    // Get Head Coord. to Virtual World transform
		    // %%% TODO re-enable this when userHeadToVworld is 
                    //     implemented correctly!!!
                    // view.getUserHeadToVworld(headToVworld);

/*******************
 %%% TODO When getUserHeadToVworld() works use code below...
                    // Get left and right ear positions in Head Coord.s
                    // Transforms left and right ears to Virtual World coord.s
                    body.getLeftEarPosition(earPosition);
                    xformLeftEar.x = (float)earPosition.x;
                    xformLeftEar.y = (float)earPosition.y;
                    xformLeftEar.z = (float)earPosition.z;
                    body.getRightEarPosition(earPosition);
                    xformRightEar.x = (float)earPosition.x;
                    xformRightEar.y = (float)earPosition.y;
                    xformRightEar.z = (float)earPosition.z;
                    // %%% TODO Put back when 
                    //      view.getUserHeadToVworld(headToVworld) can used...
                    // headToVworld.transform(xformRightEar);
                    // headToVworld.transform(xformLeftEar);
                    // Transform head viewing (Z) axis to Virtual World coord.s
                    // headToVworld.transform(xformHeadZAxis);
********************
     KLUDGE
********************/
                    Point3d  xformLeftEarD  = new Point3d();
                    Point3d  xformRightEarD = new Point3d();
                    Point3d  xformLeftEyeD  = new Point3d();
                    Point3d  xformRightEyeD = new Point3d();
                    Transform3D plateToVwrld = new Transform3D();
                    Point3d     leftEar = new Point3d();
                    Point3d     rightEar = new Point3d();
                    Point3d     leftEye = new Point3d();
                    Point3d     rightEye = new Point3d();
                    // get the calculated userHeadToVworld transform
                    // from the view cache.
                    // grab the first canvas
                    // %%% TODO not sure what to do for multiple canvases
                    Canvas3D canvas = (Canvas3D) view.getCanvas3D(0);
                    // %%% QUESTION are these methods syncronized
                    canvas.getImagePlateToVworld(plateToVwrld);
                    if (debugFlag) {
                        debugPrint("plateToVwrld colum-major:");
                        double[] matrix = new double[16];
                        plateToVwrld.get(matrix);
                        debugPrint("JSPosSample    " + matrix[0]+", " +
                             matrix[1]+", "+matrix[2]+", "+matrix[3]);
                        debugPrint("JSPosSample    " + matrix[4]+", " +
                             matrix[5]+", "+matrix[6]+", "+matrix[7]);
                        debugPrint("JSPosSample    " + matrix[8]+", " +
                             matrix[9]+", "+matrix[10]+", "+matrix[11]);
                        debugPrint("JSPosSample    " + matrix[12]+", " +
                             matrix[13]+", "+matrix[14]+", "+matrix[15]);
                    }

                    canvas.getLeftEyeInImagePlate(xformLeftEyeD);
                    canvas.getRightEyeInImagePlate(xformRightEyeD);
                    if (debugFlag) {
                        debugPrint("JSPosSample: leftEye transformed:");
                        debugPrint("                    " + xformLeftEyeD.x + ", "
                            + xformLeftEyeD.y + ", "+ xformLeftEyeD.z );
                        debugPrint("JSPosSample: rightEye transformed:");
                        debugPrint("                    " + xformRightEyeD.x +", "
                            + xformRightEyeD.y + ", "+ xformRightEyeD.z );
                    }
        
                    // Get left and right ear positions in Head Coord.s
                    // Transforms left and right ears to Virtual World coord.s
                    body.getLeftEarPosition(leftEar);
                    body.getRightEarPosition(rightEar);
                    if (debugFlag) {
                        debugPrint("JSPosSample: leftEar untransformed:");
                        debugPrint("                    " + leftEar.x + ", "
                            + leftEar.y + ", "+ leftEar.z );
                        debugPrint("JSPosSample: rightEar untransformed:");
                        debugPrint("                    " + rightEar.x + ", "
                            + rightEar.y + ", "+ rightEar.z );
                    }
                    body.getLeftEyePosition(leftEye);
                    body.getRightEyePosition(rightEye);
                    if (debugFlag) {
                        debugPrint("JSPosSample: leftEye untransformed:");
                        debugPrint("                    " + leftEye.x + ", "
                            + leftEye.y + ", "+ leftEye.z );
                        debugPrint("JSPosSample: rightEye untransformed:");
                        debugPrint("                    " + rightEye.x + ", "
                            + rightEye.y + ", "+ rightEye.z );
                    }

                    // leftEye->leftEar and rightEye->rightEar
                    // translate Eyes in ImagePlate space by offset between
                    // ears and eyes in PhysicalEnvironment space.
                    Point3d lOffset = new Point3d();
                    Point3d rOffset = new Point3d();
                    // left and right Eye in ImagePlates could be equal
                    // due to projection type, if so offset from centerEar
                    if (xformLeftEyeD.equals(xformRightEyeD)) {
                        Point3d centerEye = new Point3d();
                        centerEye.add(leftEye, rightEye);
                        centerEye.scale(0.5);
                        lOffset.sub(centerEye, leftEar);
                        rOffset.sub(centerEye, rightEar);
                    }
                    else {
                        lOffset.sub(leftEye, leftEar);
                        rOffset.sub(rightEye, rightEar);
                    }
                    if (debugFlag) {
                        debugPrint("JSPosSample: lOffset untransformed:");
                        debugPrint("                    " + lOffset.x + ", "
                            + lOffset.y + ", "+ lOffset.z );
                        debugPrint("JSPosSample: rOffset untransformed:");
                        debugPrint("                    " + rOffset.x + ", "
                            + rOffset.y + ", "+ rOffset.z );
                    }
                    // translate ears by offset for ears to eyes
                    xformLeftEarD.sub(xformLeftEyeD, lOffset);
                    xformRightEarD.sub(xformRightEyeD, rOffset);

                    // tranform Ear in ImagePlate space to VirtualWorld space
                    plateToVwrld.transform(xformLeftEarD);
                    plateToVwrld.transform(xformRightEarD);
                    if (debugFlag) {
                        debugPrint("JSPosSample: LeftEar transformed:");
                        debugPrint("                    " + xformLeftEarD.x + ", "
                            + xformLeftEarD.y + ", "+ xformLeftEarD.z );
                    }
                    xformLeftEar.set(xformLeftEarD);
                    xformRightEar.set(xformRightEarD);

                    if (lastTimeProcessed > 0) {
                        // copy old xformCenterEar into lastCenterEar
                        lastXformCenterEar.x = xformCenterEar.x;
                        lastXformCenterEar.y = xformCenterEar.y;
                        lastXformCenterEar.z = xformCenterEar.z;
                        if (debugFlag) {
                            debugPrint("           lastXformCenterEar = " + 
                                lastXformCenterEar.x + ", " + 
                                lastXformCenterEar.y + ", " + 
                                lastXformCenterEar.z);
                        }
                    }
                    // calculate the new (current) mid-point between the ears
                    // find the mid point between left and right ear positions
                    xformCenterEar.x = xformLeftEar.x + 
                         ((xformRightEar.x - xformLeftEar.x)*0.5f);
                    xformCenterEar.y = xformLeftEar.y + 
                         ((xformRightEar.y - xformLeftEar.y)*0.5f);
                    xformCenterEar.z = xformLeftEar.z + 
                         ((xformRightEar.z - xformLeftEar.z)*0.5f);
                    //%%% TODO when head changes earDirty should be set!
                    // earDirty = false;
                    if (debugFlag) {
                        debugPrint("           earXformed CALCULATED");
                        debugPrint("           xformCenterEar = " + 
                                    xformCenterEar.x + " " +
                                    xformCenterEar.y + " " +
                                    xformCenterEar.z ); 
                    }
                    earsXformed = true;
                } // end of body NOT null
              } // end of view NOT null
        } // end of earsDirty
        else {
            //%%% TODO  use existing transformed ear positions
            earsXformed = true;
        }

        if (!earsXformed) {
            // uses the default head position of (0.0, -0.03, 0.095)
            if (debugFlag)
                debugPrint("           earXformed NOT calculated");
        } 
    }

    /**  
     * Render this sample
     *
     * Calculate the audiodevice parameters necessary to spatially play this
     * sound.
     */  
    public void render(int dirtyFlags, View view, AuralParameters attribs) {
        if (debugFlag)
            debugPrint("JSPositionalSample.render"); 
        updateEar(dirtyFlags, view);

        /*
         * Time to check velocities and change the playback rate if necessary...
         *
         * Rolloff value MUST be > zero for any sound to be heard!
         *     If rolloff is zero, all sounds affected by AuralAttribute region
         *     are silent.
         * FrequencyScaleFactor value MUST be > zero for any sound to be heard!
         *     since Sfreq' = Sfreq * frequencyScaleFactor.
         *     If FrequencyScaleFactor is zero, all sounds affected by 
         *     AuralAttribute region are paused.
         * VelocityScaleFactor value of zero disables Doppler calculations.
         *
         * Scale 'Doppler' rate (or lack of Doppler) by frequencyScaleFactor.
         */
        float dopplerRatio = 1.0f;
        if (attribs != null) {
            float rolloff = attribs.getRolloff();
            float frequencyScaleFactor = attribs.getFrequencyScaleFactor();
            float velocityScaleFactor = attribs.getVelocityScaleFactor();
            if (rolloff <= 0.0f) {
                // %%% TODO Make sound silent
                // return ???
            }
            else if (frequencyScaleFactor <= 0.0f) {
                // %%% TODO Pause sound silent
                // return ???
            }
            else if (velocityScaleFactor > 0.0f) {
                if (deltaTime[curIndex] > 0) {
                    // Doppler can be calculated after the second time
                    // updateXformParams() is executed
                    dopplerRatio = calculateDoppler(attribs);

                    if (dopplerRatio > 0.0f) {
                        // rate could be changed
                        rateRatio = dopplerRatio * frequencyScaleFactor;
                        if (debugFlag) {
                            debugPrint("    scaled by frequencyScaleFactor = " +
                                    frequencyScaleFactor );
                        }
                    }
                    else if (dopplerRatio < 0.0f) {
                        // error returned by calculateDoppler
                        if (debugFlag) {
                            debugPrint("JSPositionalSample: render: " +
                                       "dopplerRatio returned = " + 
                                       dopplerRatio + "< 0");
                        }
                        // %%% TODO Make sound silent
                        // return ???
                    }
                    else {
                        // otherwise (for dopplerRatio zero) no changed
                        // %%% TODO But what if frequencyScaleFactor has changed
                        if (debugFlag) {
                            debugPrint("JSPositionalSample: render: " +
                                       "dopplerRatio returned zero; no change");
                        }
                    }
                }
            }
            else  { // auralAttributes not null but velocityFactor <= 0
              // Doppler is disabled
              rateRatio = frequencyScaleFactor; 
            }
        }
        /* 
         * since aural attributes undefined, default values are used,
         *   thus no Doppler calculated
         */
        else {
            rateRatio = 1.0f;
        }

        this.panSample(attribs);
    }    

    /* *****************
     *   
     *  Calculate Angular Gain
     *   
     * *****************/
    /*   
     *  Calculates the Gain scale factor applied to the overall gain for
     *  a sound based on angle between a sound's projected direction and the
     *  vector between the sounds position and center ear.
     *
     *  For Point Sounds this value is always 1.0f.
     */  
    float calculateAngularGain() {
        return(1.0f);
    }

    /* *****************
     *   
     *  Calculate Filter
     *   
     * *****************/
    /*   
     *  Calculates the low-pass cutoff frequency filter value applied to the 
     *  a sound based on both:
     *      Distance Filter (from Aural Attributes) based on distance
     *         between the sound and the listeners position
     *      Angular Filter (for Directional Sounds) based on the angle
     *         between a sound's projected direction and the
     *         vector between the sounds position and center ear.
     *  The lowest of these two filter is used.
     *  This filter value is stored into the sample's filterFreq field.
     */  
    void calculateFilter(float distance, AuralParameters attribs) {
        // setting filter cutoff freq to 44.1kHz which, in this
        // implementation, is the same as not performing filtering
        float   distanceFilter = 44100.0f;
        float   angularFilter  = 44100.0f;
        int arrayLength = attribs.getDistanceFilterLength();
        int filterType = attribs.getDistanceFilterType();
        boolean distanceFilterFound = false;
        boolean angularFilterFound = false;
        if ((filterType != AuralParameters.NO_FILTERING) && arrayLength > 0) {
            double[] distanceArray = new double[arrayLength];
            float[]  cutoffArray = new float[arrayLength];
            attribs.getDistanceFilter(distanceArray, cutoffArray);

            if (debugFlag) {
                debugPrint("distanceArray    cutoffArray");
                for (int i=0; i<arrayLength; i++)
                    debugPrint((float)(distanceArray[i]) + ", " + cutoffArray[i]);
            }

            distanceFilter = findFactor((double)distance, 
                   distanceArray, cutoffArray);
            if (distanceFilter < 0.0f)
                distanceFilterFound = false;
            else
                distanceFilterFound = true;
        }
        else {
            distanceFilterFound = false;
            distanceFilter = -1.0f;
        }

        if (debugFlag)
            debugPrint("    calculateFilter arrayLength = " + arrayLength);

        // Angular filter only applies to directional sound sources.
        angularFilterFound = false;
        angularFilter = -1.0f;

        filterFlag = distanceFilterFound || angularFilterFound;
        filterFreq = distanceFilter;
        if (debugFlag) 
            debugPrint("    calculateFilter flag,freq = " + filterFlag +
           "," + filterFreq );
    }

    /* *****************
     *   
     *  Find Factor
     *   
     * *****************/
    /*
     *  Interpolates the correct output factor given a 'distance' value
     *  and references to the distance array and factor array used in
     *  the calculation.  These array parameters could be either linear or
     *  angular distance arrays, or filter arrays.
     *  The values in the distance array are monotonically increasing.
     *  This method looks at pairs of distance array values to find which
     *  pair the input distance argument is between distanceArray[index] and
     *  distanceArray[index+1].
     *  The index is used to get factorArray[index] and factorArray[index+1].
     *  Then the ratio of the 'distance' between this pair of distanceArray 
     *  values is used to scale the two found factorArray values proportionally.
     *  The resulting factor is returned, unless there is an error, then -1.0
     *  is returned.
     */  
    float findFactor(double distance, 
                     double[] distanceArray, float[] factorArray) {
        int     index, lowIndex, highIndex, indexMid;

        if (debugFlag)
            debugPrint("JSPositionalSample.findFactor entered");

        /*
         * Error checking
         */
        if (distanceArray == null || factorArray == null) {
            if (debugFlag)
                debugPrint("   findFactor: arrays null");
            return -1.0f;  // no value
        }
        int arrayLength = distanceArray.length;
        if (arrayLength < 2) {
            if (debugFlag)
                debugPrint("   findFactor: arrays length < 2");
            return -1.0f; // no value
        }
        int largestIndex = arrayLength - 1;

        /*
         * Calculate distanceGain scale factor
         */
        if (distance >= distanceArray[largestIndex]) {
            if (debugFlag) {
                debugPrint("   findFactor: distance > " + 
                                  distanceArray[largestIndex]);
                debugPrint("   distanceArray length = "+ arrayLength);
            }
            return factorArray[largestIndex];
        }
        else if (distance <= distanceArray[0]) {
            if (debugFlag)
                debugPrint("   findFactor: distance < " +
                                    distanceArray[0]);
            return factorArray[0];
        } 
        /*
         * Distance between points within attenuation array.
         * Use binary halfing of distance array
         */
        else {
            lowIndex = 0;
            highIndex = largestIndex;
            if (debugFlag)
                debugPrint("   while loop to find index: ");
            while (lowIndex < (highIndex-1)) {
                if (debugFlag) {
                    debugPrint("       lowIndex " + lowIndex + 
                       ", highIndex " + highIndex);
                    debugPrint("       d.A. pair for lowIndex " + 
                       distanceArray[lowIndex] +  ", " + factorArray[lowIndex] );
                    debugPrint("       d.A. pair for highIndex " + 
                       distanceArray[highIndex] +  ", " + factorArray[highIndex] );
                }
                /*
                 * we can assume distance is between distance atttenuation vals
                 * distanceArray[lowIndex] and distanceArray[highIndex]
                 * calculate gain scale factor based on distance
                 */
                if (distanceArray[lowIndex] >= distance) {
                    if (distance < distanceArray[lowIndex]) {
                        debugPrint(true, "Internal Error: binary halving in " +
                            " findFactor failed; distance < index value");
                    }
                    if (debugFlag) {
                        debugPrint( "       index == distanceGain " +
                           lowIndex);
                        debugPrint("        findFactor returns [LOW=" + 
                           lowIndex + "] " + factorArray[lowIndex]);
                    }
                    // take value of scale factor directly from factorArray
                    return factorArray[lowIndex];
                }
                else if (distanceArray[highIndex] <= distance) {
                    if (distance > distanceArray[highIndex]) {
                        debugPrint(true, "Internal Error: binary halving in " +
                            " findFactor failed; distance > index value");
                    }
                    if (debugFlag) {
                        debugPrint( "       index == distanceGain " +
                           highIndex);
                        debugPrint("        findFactor returns [HIGH=" + 
                           highIndex + "] " + factorArray[highIndex]);
                    }
                    // take value of scale factor directly from factorArray
                    return factorArray[highIndex];
                }
                if (distance > distanceArray[lowIndex] && 
                    distance < distanceArray[highIndex] ) {
                    indexMid = lowIndex + ((highIndex - lowIndex) / 2);
                    if (distance <= distanceArray[indexMid])
                        // value of distance in lower "half" of list
                        highIndex = indexMid;
                    else // value if distance in upper "half" of list
                        lowIndex = indexMid;
                }
            } /* of while */

            /*
             * ratio: distance from listener to sound source 
             *        between lowIndex and highIndex times
             *        attenuation value between lowIndex and highIndex
             * gives linearly interpolationed attenuation value
             */
            if (debugFlag) {
                debugPrint( "   ratio calculated using lowIndex " +
                       lowIndex + ", highIndex " + highIndex);
                debugPrint( "   d.A. pair for lowIndex " + 
                        distanceArray[lowIndex]+", "+factorArray[lowIndex] );
                debugPrint( "   d.A. pair for highIndex " + 
                        distanceArray[highIndex]+", "+factorArray[highIndex] );
            }

            float outputFactor = 
                   ((float)(((distance - distanceArray[lowIndex])/
                    (distanceArray[highIndex] - distanceArray[lowIndex]) ) ) *
                    (factorArray[highIndex] - factorArray[lowIndex]) ) +
                   factorArray[lowIndex] ; 
            if (debugFlag)
                debugPrint("    findFactor returns " + outputFactor);
            return outputFactor;
        }  
    }

    /**
     * CalculateDistanceAttenuation
     *
     * Simply calls generic (for PointSound) 'findFactor()' with 
     * a single set of attenuation distance and gain scale factor arrays.
     */
    float calculateDistanceAttenuation(float distance) {
        float  factor = 1.0f;
        factor = findFactor((double)distance, this.attenuationDistance,
                           this.attenuationGain);
        if (factor >= 0.0)
            return (factor);
        else
            return (1.0f);
    }

    /* ******************
     *   
     *  Pan Sample
     *   
     * ******************/
    /*  
     *  Sets pan and delay for a single sample associated with this Sound.
     *  Front and Back quadrants are treated the same.
     */
    void panSample(AuralParameters attribs) {
        int     quadrant = 1;
        float   intensityHigh = 1.0f;
        float   intensityLow = 0.125f;
        float   intensityDifference = intensityHigh - intensityLow;

        //&&& time around "average" default head
        //&&& int     delayHigh = 32;  // 32.15 samples = .731 ms
        //&&& int     delayLow = 0;

        boolean front;
        float   intensityOffset; // 0.0 -> 1.0 then 1.0 -> 0.0 for full rotation
        float   halfX;
        int     id;
        int     err;

        float   nearZero = 0.000001f;
        float   nearOne  = 0.999999f;
        float   nearNegativeOne  = -nearOne;
        float   halfPi = (float)Math.PI * 0.5f;
        /*
         * Parameters used for IID and ITD equations.
         * Name of parameters (as used in Guide, E.3) are denoted in comments.
         */
        float    distanceSourceToCenterEar = 0.0f;     // Dh
        float    lastDistanceSourceToCenterEar = 0.0f;
        float    distanceSourceToRightEar = 0.0f;      // Ef or Ec
        float    distanceSourceToLeftEar = 0.0f;       // Ef or Ec
        float    distanceBetweenEars = 0.18f;          // De
        float    radiusOfHead = 0.0f;                  // De/2
        float    radiusOverDistanceToSource = 0.0f;    // De/2 * 1/Dh

        float    alpha = 0.0f;                         // 'alpha'
        float    sinAlpha = 0.0f;                      // sin(alpha);
        float    gamma = 0.0f;                         // 'gamma'

        // Speed of Sound (unaffected by rolloff) in millisec/meters
        float    speedOfSound = attribs.SPEED_OF_SOUND;
        float    invSpeedOfSound = 1.0f / attribs.SPEED_OF_SOUND;

        float    sampleRate = 44.1f;                   // 44 samples/millisec

        boolean  rightEarClosest = false;
        boolean  soundFromBehind = false;

        float    distanceGain = 1.0f;
        float    allGains = this.gain; // product of gain scale factors
  
        Point3f  workingPosition = new Point3f();
        Transform3D headToVworld = new Transform3D();

        // Asuumes that head and ear positions can be retrieved from universe

        Vector3f mixScale = new Vector3f();  // for mix*Samples code

        // Use transformed position of this sound
        workingPosition.set(this.xformPosition);
        if (debugFlag)
            debugPrint("panSample:workingPosition from xformed point " +
                 this.xformPosition.x + " " + this.xformPosition.y + " " +
                 this.xformPosition.z + " for pointSound " + this);

        // Create the vectors from the sound source to head positions
        sourceToCenterEar.x = xformCenterEar.x - workingPosition.x;
        sourceToCenterEar.y = xformCenterEar.y - workingPosition.y;
        sourceToCenterEar.z = xformCenterEar.z - workingPosition.z;
        sourceToRightEar.x = xformRightEar.x - workingPosition.x;
        sourceToRightEar.y = xformRightEar.y - workingPosition.y;
        sourceToRightEar.z = xformRightEar.z - workingPosition.z;
        sourceToLeftEar.x = xformLeftEar.x - workingPosition.x;
        sourceToLeftEar.y = xformLeftEar.y - workingPosition.y;
        sourceToLeftEar.z = xformLeftEar.z - workingPosition.z;

        /*
         * get distances from SoundSource to 
         *    (i)   head origin
         *    (ii)  right ear
         *    (iii) left ear
         */
        distanceSourceToCenterEar = workingPosition.distance(xformCenterEar);
        distanceSourceToRightEar = workingPosition.distance(xformRightEar);
        distanceSourceToLeftEar = workingPosition.distance(xformLeftEar);
        distanceBetweenEars = xformRightEar.distance(xformLeftEar);
        if (debugFlag)
            debugPrint("           distance from left,right ears to source: = (" +
                  distanceSourceToLeftEar + ", " + distanceSourceToRightEar + ")");

        radiusOfHead = distanceBetweenEars * 0.5f;
        if (debugFlag)
            debugPrint("           radius of head = " + radiusOfHead );
        radiusOverDistanceToSource =    // De/2 * 1/Dh
                    radiusOfHead/distanceSourceToCenterEar;
        if (debugFlag)
            debugPrint("           radius over distance = " + radiusOverDistanceToSource );
        // Dot Product 
        double dotProduct = (double)(
                    (sourceToCenterEar.dot(xformHeadZAxis))/
                    (sourceToCenterEar.length() * xformHeadZAxis.length()));
        if (debugFlag)
            debugPrint( "           dot product = " + dotProduct );
        alpha = (float)(Math.acos(dotProduct));
        if (debugFlag)
            debugPrint( "           alpha = " + alpha );

        if (alpha > halfPi) {
            if (debugFlag)
                debugPrint("           sound from behind");
            soundFromBehind = true;
            alpha = (float)Math.PI - alpha;
            if (debugFlag)
                debugPrint( "           PI minus alpha =>" + alpha );
        }
        else {
            soundFromBehind = false;
            if (debugFlag)
                debugPrint("           sound from in front");
        }

        gamma = (float)(Math.acos(radiusOverDistanceToSource));
        if (debugFlag)
            debugPrint( "           gamma " + gamma );

        rightEarClosest = 
            (distanceSourceToRightEar>distanceSourceToLeftEar) ? false : true ;
        if (rightEarClosest)
            if (debugFlag)
                debugPrint( "           right ear closest");
        else 
            if (debugFlag)
                debugPrint( "           left ear closest");
        sinAlpha = (float)(Math.sin((double)alpha));
        if (sinAlpha < 0.0) sinAlpha = -sinAlpha;
        if (debugFlag)
            debugPrint( "           sin(alpha) " + sinAlpha );

        /*
         * The path from sound source to the farthest ear is always indirect
         * (it wraps around part of the head).
         * Calculate distance wrapped around the head for farthest ear
         */
        float DISTANCE = (float)Math.sqrt((double)
                distanceSourceToCenterEar * distanceSourceToCenterEar +
                radiusOfHead * radiusOfHead);
        if (debugFlag)
            debugPrint( "           partial distance from edge of head to source = "
                      + distanceSourceToCenterEar);
        if (rightEarClosest) {
            distanceSourceToLeftEar = 
                DISTANCE + radiusOfHead * (halfPi+alpha-gamma);
            if (debugFlag)
                debugPrint("           new distance from left ear to source = "
                      + distanceSourceToLeftEar);
        } 
        else {
            distanceSourceToRightEar = 
                DISTANCE + radiusOfHead * (halfPi+alpha-gamma);
            if (debugFlag)
                debugPrint("           new distance from right ear to source = "
                      + distanceSourceToRightEar);
        }
        /*
         * The path from the source source to the closest ear could either
         * be direct or indirect (wraps around part of the head).
         * if sinAlpha >= radiusOverDistance path of sound to closest ear
         * is direct, otherwise it is indirect
         */
        if (sinAlpha < radiusOverDistanceToSource) {
            if (debugFlag)
                debugPrint("           closest path is also indirect ");
            // Path of sound to closest ear is indirect
 
            if (rightEarClosest) {
                distanceSourceToRightEar = 
                    DISTANCE + radiusOfHead * (halfPi-alpha-gamma);
                if (debugFlag)
                    debugPrint("           new distance from right ear to source = "
                      + distanceSourceToRightEar);
            }
            else {
                distanceSourceToLeftEar = 
                    DISTANCE + radiusOfHead * (halfPi-alpha-gamma);
                if (debugFlag)
                    debugPrint("           new distance from left ear to source = "
                      + distanceSourceToLeftEar);
            }
        }
        else {
            if (debugFlag)
                debugPrint("           closest path is direct ");
            if (rightEarClosest) {
                if (debugFlag)
                    debugPrint("           direct distance from right ear to source = "
                      + distanceSourceToRightEar);
            }
            else {
                if (debugFlag)
                    debugPrint("           direct distance from left ear to source = "
                      + distanceSourceToLeftEar);
            }
        }

        /**
         * Short-cut taken.  Rather than using actual delays from source
         * (where the overall distances would be taken into account in 
         * determining delay) the difference in the left and right delay
         * are applied.
         * This approach will be preceptibly wrong for sound sources that
         * are very far away from the listener so both ears would have 
         * large delay.
         */
        if (rightEarClosest) {
            rightDelay = 0;
            leftDelay = (int)((distanceSourceToLeftEar - distanceSourceToRightEar) *
                    invSpeedOfSound * sampleRate);
        }
        else {
            leftDelay = 0;
            rightDelay = (int)((distanceSourceToRightEar - distanceSourceToLeftEar) *
                    invSpeedOfSound * sampleRate);
        }

        if (debugFlag)
            debugPrint("           left and right delay = ("
                      + leftDelay + ", " + rightDelay + ")");

        // What should the gain be for the different ears???
        //%%% TODO now using a Kludge that sets gain based on a unit circle!!!
        workingPosition.sub(xformCenterEar); // offset sound pos. by head origin
        // normalize; put Sound on unit sphere around head origin
        workingPosition.scale(1.0f/distanceSourceToCenterEar);
        if (debugFlag)
            debugPrint("           workingPosition after unitization " +
              workingPosition.x+" "+workingPosition.y+" "+workingPosition.z );

        /*
         * Get the correct distance gain scale factor from attenuation arrays.
         * This requires that sourceToCenterEar vector has been calculated.
         */
        //%%% TODO now using distance from center ear to source
        //    Using distances from each ear to source would be more accurate
        distanceGain = calculateDistanceAttenuation(distanceSourceToCenterEar);

        allGains *= distanceGain;

        /*
         * Add angular gain (for Cone sound)
         */
        if (debugFlag)
            debugPrint("           all Gains (without angular gain) " + allGains);
        // assume that xformPosition is already calculated
        allGains *= this.calculateAngularGain();
        if (debugFlag)
            debugPrint("                     (incl. angular gain)  " + allGains);

        halfX = workingPosition.x/2.0f;
        if (halfX >= 0)
            intensityOffset = (intensityDifference * (0.5f - halfX));
        else
            intensityOffset = (intensityDifference * (0.5f + halfX));

        /*
         * Determine the quadrant sound is in 
         */
        if (workingPosition.x >= 0.0) {
            front = false;
            if (workingPosition.z >= 0)
                quadrant = 1;
            else
                quadrant = 4;
        }
        else { // workingPosition.x < 0.0 
            front = false;
            if (workingPosition.z >= 0)
                quadrant = 2;
            else
                quadrant = 3;
        }
        /*
         * For now have delay constant for front back sound for now
         */
        if (debugFlag)
            debugPrint("panSample:                 quadrant " + quadrant);
        switch (quadrant) {
          case 1: 
            /*
             * X is positive, from 1.0 to 0.0
             * Sound from front
             */
          case 4:
            /*
             * X is positive, from 0.0 to 1.0
             * Sound from back
             */
            rightGain = allGains * (intensityHigh - intensityOffset);
            leftGain =  allGains * (intensityLow + intensityOffset);
            break;

          case 2:
            /*
             * X is negitive, from 0.0 to -1.0
             * Sound from front
             */
          case 3:
            /*
             * X is negitive, from -1.0 to 0.0
             * Sound from back
             */
            leftGain = allGains * (intensityHigh - intensityOffset);
            rightGain = allGains * (intensityLow + intensityOffset);
            break;
        }  /* switch */
        if (debugFlag)
            debugPrint("panSample:                 left/rightGain " + leftGain +
                        ", " + rightGain);

        // Combines distance and angular filter to set this sample's current
        // frequency cutoff value
        calculateFilter(distanceSourceToCenterEar, attribs); 

    }  /* panSample() */

    /*
     * Override stopSample
     * Reset feilds necessary for future start
     */  
    public void stop() {
        super.stop();
        if (debugFlag)
            debugPrint("JSPosSample.stop() resets lastTimeProcessed to -1");
        lastTimeProcessed = -1;  // denotes not previously processed
    }

}
